Imported Upstream version 2.12.5
[platform/upstream/git.git] / attr.c
diff --git a/attr.c b/attr.c
index 734222d..1fcf042 100644 (file)
--- a/attr.c
+++ b/attr.c
@@ -12,6 +12,7 @@
 #include "exec_cmd.h"
 #include "attr.h"
 #include "dir.h"
+#include "utf8.h"
 
 const char git_attr__true[] = "(builtin)true";
 const char git_attr__false[] = "\0(builtin)false";
@@ -32,9 +33,12 @@ struct git_attr {
        struct git_attr *next;
        unsigned h;
        int attr_nr;
+       int maybe_macro;
+       int maybe_real;
        char name[FLEX_ARRAY];
 };
 static int attr_nr;
+static int cannot_trust_maybe_real;
 
 static struct git_attr_check *check_all_attr;
 static struct git_attr *(git_attr_hash[HASHSIZE]);
@@ -89,16 +93,15 @@ static struct git_attr *git_attr_internal(const char *name, int len)
        if (invalid_attr_name(name, len))
                return NULL;
 
-       a = xmalloc(sizeof(*a) + len + 1);
-       memcpy(a->name, name, len);
-       a->name[len] = 0;
+       FLEX_ALLOC_MEM(a, name, name, len);
        a->h = hval;
        a->next = git_attr_hash[pos];
        a->attr_nr = attr_nr++;
+       a->maybe_macro = 0;
+       a->maybe_real = 0;
        git_attr_hash[pos] = a;
 
-       check_all_attr = xrealloc(check_all_attr,
-                                 sizeof(*check_all_attr) * attr_nr);
+       REALLOC_ARRAY(check_all_attr, attr_nr);
        check_all_attr[a->attr_nr].attr = a;
        check_all_attr[a->attr_nr].value = ATTR__UNKNOWN;
        return a;
@@ -119,7 +122,7 @@ struct pattern {
        const char *pattern;
        int patternlen;
        int nowildcardlen;
-       int flags;              /* EXC_FLAG_* */
+       unsigned flags;         /* EXC_FLAG_* */
 };
 
 /*
@@ -245,9 +248,10 @@ static struct match_attr *parse_attr_line(const char *line, const char *src,
                      sizeof(*res) +
                      sizeof(struct attr_state) * num_attr +
                      (is_macro ? 0 : namelen + 1));
-       if (is_macro)
+       if (is_macro) {
                res->u.attr = git_attr_internal(name, namelen);
-       else {
+               res->u.attr->maybe_macro = 1;
+       } else {
                char *p = (char *)&(res->state[num_attr]);
                memcpy(p, name, namelen);
                res->u.pat.pattern = p;
@@ -267,6 +271,10 @@ static struct match_attr *parse_attr_line(const char *line, const char *src,
        /* Second pass to fill the attr_states */
        for (cp = states, i = 0; *cp; i++) {
                cp = parse_attr(src, lineno, cp, &(res->state[i]));
+               if (!is_macro)
+                       res->state[i].attr->maybe_real = 1;
+               if (res->state[i].attr->maybe_macro)
+                       cannot_trust_maybe_real = 1;
        }
 
        return res;
@@ -370,8 +378,12 @@ static struct attr_stack *read_attr_from_file(const char *path, int macro_ok)
                return NULL;
        }
        res = xcalloc(1, sizeof(*res));
-       while (fgets(buf, sizeof(buf), fp))
-               handle_attr_line(res, buf, path, ++lineno, macro_ok);
+       while (fgets(buf, sizeof(buf), fp)) {
+               char *bufp = buf;
+               if (!lineno)
+                       skip_utf8_bom(&bufp, strlen(bufp));
+               handle_attr_line(res, bufp, path, ++lineno, macro_ok);
+       }
        fclose(fp);
        return res;
 }
@@ -476,10 +488,11 @@ static int git_attr_system(void)
        return !git_env_bool("GIT_ATTR_NOSYSTEM", 0);
 }
 
+static GIT_PATH_FUNC(git_path_info_attributes, INFOATTRIBUTES_FILE)
+
 static void bootstrap_attr_stack(void)
 {
        struct attr_stack *elem;
-       char *xdg_attributes_file;
 
        if (attr_stack)
                return;
@@ -498,10 +511,8 @@ static void bootstrap_attr_stack(void)
                }
        }
 
-       if (!git_attributes_file) {
-               home_config_paths(NULL, &xdg_attributes_file, "attributes");
-               git_attributes_file = xdg_attributes_file;
-       }
+       if (!git_attributes_file)
+               git_attributes_file = xdg_config_home("attributes");
        if (git_attributes_file) {
                elem = read_attr_from_file(git_attributes_file, 1);
                if (elem) {
@@ -520,7 +531,11 @@ static void bootstrap_attr_stack(void)
                debug_push(elem);
        }
 
-       elem = read_attr_from_file(git_path(INFOATTRIBUTES_FILE), 1);
+       if (startup_info->have_repository)
+               elem = read_attr_from_file(git_path_info_attributes(), 1);
+       else
+               elem = NULL;
+
        if (!elem)
                elem = xcalloc(1, sizeof(*elem));
        elem->origin = NULL;
@@ -682,13 +697,14 @@ static int fill(const char *path, int pathlen, int basename_offset,
        return rem;
 }
 
-static int macroexpand_one(int attr_nr, int rem)
+static int macroexpand_one(int nr, int rem)
 {
        struct attr_stack *stk;
        struct match_attr *a = NULL;
        int i;
 
-       if (check_all_attr[attr_nr].value != ATTR__TRUE)
+       if (check_all_attr[nr].value != ATTR__TRUE ||
+           !check_all_attr[nr].attr->maybe_macro)
                return rem;
 
        for (stk = attr_stack; !a && stk; stk = stk->prev)
@@ -696,7 +712,7 @@ static int macroexpand_one(int attr_nr, int rem)
                        struct match_attr *ma = stk->attrs[i];
                        if (!ma->is_macro)
                                continue;
-                       if (ma->u.attr->attr_nr == attr_nr)
+                       if (ma->u.attr->attr_nr == nr)
                                a = ma;
                }
 
@@ -707,10 +723,13 @@ static int macroexpand_one(int attr_nr, int rem)
 }
 
 /*
* Collect all attributes for path into the array pointed to by
* check_all_attr.
  */
-static void collect_all_attrs(const char *path)
+static void collect_some_attrs(const char *path, int num,
+                              struct git_attr_check *check)
+
 {
        struct attr_stack *stk;
        int i, pathlen, rem, dirlen;
@@ -733,6 +752,19 @@ static void collect_all_attrs(const char *path)
        prepare_attr_stack(path, dirlen);
        for (i = 0; i < attr_nr; i++)
                check_all_attr[i].value = ATTR__UNKNOWN;
+       if (num && !cannot_trust_maybe_real) {
+               rem = 0;
+               for (i = 0; i < num; i++) {
+                       if (!check[i].attr->maybe_real) {
+                               struct git_attr_check *c;
+                               c = check_all_attr + check[i].attr->attr_nr;
+                               c->value = ATTR__UNSET;
+                               rem++;
+                       }
+               }
+               if (rem == num)
+                       return;
+       }
 
        rem = attr_nr;
        for (stk = attr_stack; 0 < rem && stk; stk = stk->prev)
@@ -743,7 +775,7 @@ int git_check_attr(const char *path, int num, struct git_attr_check *check)
 {
        int i;
 
-       collect_all_attrs(path);
+       collect_some_attrs(path, num, check);
 
        for (i = 0; i < num; i++) {
                const char *value = check_all_attr[check[i].attr->attr_nr].value;
@@ -759,7 +791,7 @@ int git_all_attrs(const char *path, int *num, struct git_attr_check **check)
 {
        int i, count, j;
 
-       collect_all_attrs(path);
+       collect_some_attrs(path, 0, NULL);
 
        /* Count the number of attributes that are set. */
        count = 0;
@@ -769,7 +801,7 @@ int git_all_attrs(const char *path, int *num, struct git_attr_check **check)
                        ++count;
        }
        *num = count;
-       *check = xmalloc(sizeof(**check) * count);
+       ALLOC_ARRAY(*check, count);
        j = 0;
        for (i = 0; i < attr_nr; i++) {
                const char *value = check_all_attr[i].value;
@@ -707,10 +723,13 @@ static int macroexpand_one(int attr_nr, int rem)
 
 /*
- * Collect all attributes for path into the array pointed to by
- * check_all_attr.
+ * Collect attributes for path into the array pointed to by
+ * check_all_attr. If num is non-zero, only attributes in check[] are
+ * collected. Otherwise all attributes are collected.
  */